--[[
DamageMeters.lua
See the Readme for author and permissions.


TODO:
 FEATURES
- Options menu.
- Right click on bar for delete, banish.
- bars currently pulse with any activity, not the selected quantity

 KNOWN BUGS
- Tooltips not showing up all the time, maybe after a dmhide?
- Hostile player buff (use crusader for test).
- definitely missing healing done by own party members

<Backdrop bgFile="Interface\DialogFrame\UI-DialogBox-Background" edgeFile="Interface\DialogFrame\UI-DialogBox-Border" tile="true">

]]--

-------------------------------------------------------------------------------

-- CONSTANTS ---
DamageMeters_BARCOUNT_MAX = 40;	-- Note this is also the table max size atm.
DamageMeters_BARWIDTH = 119;
DamageMeters_PULSE_TIME = 1.00;

DamageMeters_Sort_DECREASING = 1;
DamageMeters_Sort_INCREASING = 2;
DamageMeters_Sort_MAX = 2;

DamageMeters_Quantity_DAMAGE = 1;
DamageMeters_Quantity_HEALING = 2;
DamageMeters_Quantity_TIME = 3;
DamageMeters_Quantity_MAX = 3;

DamageMeters_Relation_SELF = 1;
DamageMeters_Relation_PET = 2;
DamageMeters_Relation_PARTY = 3;
DamageMeters_Relation_FRIENDLY = 4;

DamageMeters_colorScheme_RELATIONSHIP = 1;
DamageMeters_colorScheme_CLASS = 2;
DamageMeters_colorScheme_MAX = 2;

-- GLOBALS ---
DamageMeters_bars = {};
DamageMeters_text = {};
DamageMeters_table = {}; -- damage, healing, critCount, hitCount, lastTime, relationship, class
DamageMeters_savedTable = {};
DamageMeters_tooltipBarIndex = nil;

-- SETTINGS ---
DamageMeters_BarCount = 10;
DamageMeters_quantity = DamageMeters_Quantity_DAMAGE;
DamageMeters_sort = DamageMeters_Sort_DECREASING;
DamageMeters_showName = true;
DamageMeters_showPercentage = false;
DamageMeters_showValue = false;
DamageMeters_colorScheme = DamageMeters_colorScheme_CLASS;

-- DEBUG --
DamageMeters_msgCounts = {};

-------------------------------------------------------------------------------

function DMPrint(msg)
	local r = 0.50;
	local g = 0.50;
	local b = 1.00;
	if (not frame) then 
		frame = DEFAULT_CHAT_FRAME; 
	end
	if (frame) then
		frame:AddMessage(msg,r,g,b);
	end
end

function DMPrintD(msg)
	-- SendChatMessage(msg, "CHANNEL", nil, GetChannelName("dmdebug"));	
end

function DMVerbose(msg)
	--DMPrint(msg);
end

-------------------------------------------------------------------------------

function DamageMetersFrame_OnLoad()
	local name = this:GetName();
	local i;
	for i = 1,DamageMeters_BARCOUNT_MAX do
		DamageMeters_bars[i] = getglobal(name.."Bar"..i);
		DamageMeters_bars[i]:SetID(i);
		DamageMeters_bars[i]:Hide();
		DamageMeters_text[i] = getglobal(name.."Text"..i);
		SetTextStatusBarText(DamageMeters_bars[i], DamageMeters_text[i]);
		-- Force text on always.
		ShowTextStatusBarText(DamageMeters_bars[i]);
	end

	-- Apparently this only works for buttons?
	--this:RegisterForClicks("RightButtonUp");

	this:RegisterEvent("VARIABLES_LOADED");

	this:RegisterEvent("CHAT_MSG_COMBAT_SELF_HITS");	-- Melee you do on things.
	this:RegisterEvent("CHAT_MSG_COMBAT_PET_HITS");		-- Melee your pets do.
	this:RegisterEvent("CHAT_MSG_COMBAT_PARTY_HITS");	-- Melee done by part.
	this:RegisterEvent("CHAT_MSG_COMBAT_FRIENDLYPLAYER_HITS");	-- Melee done by friendlies.

	this:RegisterEvent("CHAT_MSG_SPELL_SELF_DAMAGE");	-- Your spells that damage other things.
	this:RegisterEvent("CHAT_MSG_SPELL_PET_DAMAGE");
	this:RegisterEvent("CHAT_MSG_SPELL_PARTY_DAMAGE");	-- Party member's spell hits.
	this:RegisterEvent("CHAT_MSG_SPELL_FRIENDLYPLAYER_DAMAGE"); -- Spells other people cast on things.

	this:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_CREATURE_DAMAGE"); -- Blah suffers # Arcane damage from #'s/your Spell.  Works on self, party, friendly.
	this:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_HOSTILEPLAYER_DAMAGE");

	this:RegisterEvent("CHAT_MSG_SPELL_DAMAGESHIELDS_ON_SELF");		-- Thorns on self.
	this:RegisterEvent("CHAT_MSG_SPELL_DAMAGESHIELDS_ON_OTHERS");	-- Thorns on others.

	this:RegisterEvent("CHAT_MSG_SPELL_SELF_BUFF");
	this:RegisterEvent("CHAT_MSG_SPELL_PARTY_BUFF");
	this:RegisterEvent("CHAT_MSG_SPELL_FRIENDLYPLAYER_BUFF");

	this:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS");
	this:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_PARTY_BUFFS");
	this:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_FRIENDLYPLAYER_BUFFS");

	--!
	this:RegisterEvent("CHAT_MSG_SPELL_HOSTILEPLAYER_BUFF");
	this:RegisterEvent("CHAT_MSG_SPELL_PERIODIC_HOSTILEPLAYER_BUFFS");

	SlashCmdList["DAMAGEMETERSHELP"] = DamageMeters_Help;
	SLASH_DAMAGEMETERSHELP1 = "/dmhelp";
	SlashCmdList["DAMAGEMETERSSHOW"] = DamageMeters_Show;
	SLASH_DAMAGEMETERSSHOW1 = "/dmshow";
	SlashCmdList["DAMAGEMETERSHIDE"] = DamageMeters_Hide;
	SLASH_DAMAGEMETERSHIDE1 = "/dmhide";
	SlashCmdList["DAMAGEMETERSCLEAR"] = DamageMeters_Clear;
	SLASH_DAMAGEMETERSCLEAR1 = "/dmclear";
	SlashCmdList["DAMAGEMETERSREPORT"] = DamageMeters_Report;
	SLASH_DAMAGEMETERSREPORT1 = "/dmreport";
	SlashCmdList["DAMAGEMETERSSORT"] = DamageMeters_SetSort;
	SLASH_DAMAGEMETERSSORT1 = "/dmsort";
	SlashCmdList["DAMAGEMETERSCOUNT"] = DamageMeters_SetCount;
	SLASH_DAMAGEMETERSCOUNT1 = "/dmcount";
	SlashCmdList["DAMAGEMETERSSAVE"] = DamageMeters_Save;
	SLASH_DAMAGEMETERSSAVE1 = "/dmsave";
	SlashCmdList["DAMAGEMETERSRESTORE"] = DamageMeters_Restore;
	SLASH_DAMAGEMETERSRESTORE1 = "/dmrestore";
	SlashCmdList["DAMAGEMETERSMERGE"] = DamageMeters_Merge;
	SLASH_DAMAGEMETERSMERGE1 = "/dmmerge";
	SlashCmdList["DAMAGEMETERSSWAP"] = DamageMeters_Swap;
	SLASH_DAMAGEMETERSSWAP1 = "/dmswap";
	SlashCmdList["DAMAGEMETERSRESETPOS"] = DamageMeters_ResetPos;
	SLASH_DAMAGEMETERSRESETPOS1 = "/dmresetpos";
	SlashCmdList["DAMAGEMETERSSHOWTEXT"] = DamageMeters_SetTextOptions;
	SLASH_DAMAGEMETERSSHOWTEXT1 = "/dmtext";
	SlashCmdList["DAMAGEMETERSCOLORSCHEME"] = DamageMeters_SetColorScheme;
	SLASH_DAMAGEMETERSCOLORSCHEME1 = "/dmcolor";
	SlashCmdList["DAMAGEMETERSQUANTITY"] = DamageMeters_SetQuantity;
	SLASH_DAMAGEMETERSQUANTITY1 = "/dmquant";
	SlashCmdList["DAMAGEMETERSRESET"] = DamageMeters_Reset;
	SLASH_DAMAGEMETERSRESET1 = "/dmreset";
	SlashCmdList["DAMAGEMETERSTEST"] = DamageMeters_Test;
	SLASH_DAMAGEMETERSTEST1 = "/dmtest";
	SlashCmdList["DAMAGEMETERSADD"] = DamageMeters_Add;
	SLASH_DAMAGEMETERSADD1 = "/dmadd";
	SlashCmdList["DAMAGEMETERSDUMPTABLE"] = DamageMeters_DumpTable;
	SLASH_DAMAGEMETERSDUMPTABLE1 = "/dmdumptable";

	SlashCmdList["DAMAGEMETERSDUMPMSG"] = DM_DumpMsg;
	SLASH_DAMAGEMETERSDUMPMSG1 = "/dmdumpmsg";

	RegisterForSave("DamageMeters_BarCount");
	RegisterForSave("DamageMeters_quantity");
	RegisterForSave("DamageMeters_sort");
	RegisterForSave("DamageMeters_showName");
	RegisterForSave("DamageMeters_showPercentage");
	RegisterForSave("DamageMeters_showValue");
	RegisterForSave("DamageMeters_colorScheme");

	--damageMeters_Clear();
	DamageMetersFrame_GenerateFrame(this);
end

function DamageMetersFrame_OnUpdate()
	local totalValue = 0;
	local totalDamage = 0;
	local totalHealing = 0;
	local maxUnitIndex = 0;
	local maxUnitValue = 0;
	local currentTime = GetTime();

	-- Sort the table.
	DamageMeters_DoSort();

	local index;
	local info;
	for index,info in DamageMeters_table do 
		local unitValue;
		
		if (DamageMeters_quantity == DamageMeters_Quantity_TIME) then
			unitValue = currentTime - info.lastTime;
		elseif (DamageMeters_quantity == DamageMeters_Quantity_HEALING) then
			unitValue = info.healing;
		else
			unitValue = info.damage;
		end

		if (unitValue > maxUnitValue) then
			maxUnitIndex = index;
			maxUnitValue = unitValue;
		end

		totalValue = totalValue + unitValue;
		totalHealing = totalHealing + info.healing;
		totalDamage = totalDamage + info.damage; 
	end

	local barIndex = 1;

	local i;
	for i = 1,DamageMeters_BarCount do
		DamageMeters_bars[i]:SetMinMaxValues(0, maxUnitValue);
		DamageMeters_text[i]:SetText("");
	end

	-- Tooltip
	if (DamageMetersTooltip:IsOwned(this)) then
		DamageMeters_SetTooltipText(totalDamage, totalHealing);
	end

	barIndex = 1;
	local struct;
	for i,struct in DamageMeters_table do 
		DamageMetersFrame_SetBarInfo(barIndex, totalValue, p == maxUnitIndex);

		barIndex = barIndex + 1;
		if (barIndex > DamageMeters_BarCount) then
			return;
		end
	end
end

local DamageMeters_BARHEIGHT = 12;

function DamageMetersFrame_GenerateFrame(frame)
	if (not frame) then
		frame = getglobal("DamageMetersFrame");
		if (not frame) then
			DMPrint("DamageMetersFrame_GenerateFrame : Error getting frame.");
		end
	end

	-- Hide all bars: update will reshow those that need to be seen.
	local i;
	for i = 1,DamageMeters_BARCOUNT_MAX do
		DamageMeters_bars[i]:Hide();
		DamageMeters_bars[i]:SetValue(0);
		DamageMeters_text[i]:SetText("");
	end

	local rowCount = 0;
	local columnCount = 1;
	if (DamageMeters_BarCount > (DamageMeters_BARCOUNT_MAX / 2)) then
		rowCount = DamageMeters_BarCount / 2;
		columnCount = 2;
		frame:SetWidth( DamageMeters_BARWIDTH * 2 + 10 + 2 );
	else
		columnCount = 1;
		rowCount = DamageMeters_BarCount;
		frame:SetWidth( DamageMeters_BARWIDTH + 10 );
	end
	frame:SetHeight( (DamageMeters_BARHEIGHT * rowCount) + 11 );

	--DMPrint("DamageMeters: "..rowCount.." rows, "..columnCount.." columns.");

	local name = frame:GetName();
	local row;
	local column;
	for row = 1, rowCount do
		for column = 1, columnCount do
			--DMPrint("Row = "..row..", column = "..column);
			local index = row + (column - 1) * rowCount;
			local itemButton = DamageMeters_bars[index];
			local textButton = DamageMeters_text[index];

			local x = 5 + (column - 1) * (DamageMeters_BARWIDTH + 2);
			local y = -6 - (row - 1) * DamageMeters_BARHEIGHT;
			itemButton:SetPoint("TOPLEFT", name, "TOPLEFT", x, y);
			textButton:SetPoint("CENTER", itemButton:GetName(), "CENTER", 0, 0);
		end
	end

	DamageMeters_SetBackgroundColor();
end

function DamageMetersFrame_SetBarInfo(barIndex, totalDamage, totalHealing, isMax)
	local red = 0.00;
	local green = 0.00;
	local blue = 0.00;
	local barString = "";

	local player = DamageMeters_table[barIndex].player;
	local age = GetTime() - DamageMeters_table[barIndex].lastTime;

	local isPlayer = (DamageMeters_Relation_SELF == DamageMeters_table[barIndex].relationship);

	local relationship = DamageMeters_table[barIndex].relationship;
	if (DamageMeters_colorScheme == 1) then
		if (DamageMeters_Relation_SELF == relationship) then
			green = 1.00;
		elseif (DamageMeters_Relation_PET == relationship) then
			green = 0.80;
		elseif (DamageMeters_Relation_PARTY == relationship) then
			blue = 1.00;
		elseif (DamageMeters_Relation_FRIENDLY == relationship) then
			red = 1.00;
			green = 0.50;
		end
	else
		local class = DamageMeters_table[barIndex].class;
		if (class) then
			local color = RAID_CLASS_COLORS[string.upper(class)];
			red = color.r;
			green = color.g;
			blue = color.b;
		elseif (DamageMeters_Relation_PET == relationship) then
			red = 0.00;
			green = 0.80;
			blue = 0.00;
		else
			red = 0.70;
			green = 0.70;
			blue = 0.70;
		end
	end

	local pulseMag = 0.00;
	if (age < DamageMeters_PULSE_TIME) then
		pulseMag = 1.00 - age / DamageMeters_PULSE_TIME;
	end

	local value;
	if ((DamageMeters_quantity == DamageMeters_Quantity_DAMAGE) or (DamageMeters_quantity == DamageMeters_Quantity_HEALING)) then
		if (DamageMeters_quantity == DamageMeters_Quantity_DAMAGE) then
			value = DamageMeters_table[barIndex].damage;
		else
			value = DamageMeters_table[barIndex].healing;
		end

		if (DamageMeters_showName) then
			barString = player;
		end
		if (DamageMeters_showPercentage) then
			local percent = (totalDamage > 0) and ((value / totalDamage) * 100) or 0;
			barString = format("%s %.2f%%", barString, percent);
		end
		if (DamageMeters_showValue) then
			barString = format("%s %d", barString, value);
		end	
	elseif (DamageMeters_quantity == DamageMeters_Quantity_TIME) then
		value = age;
		barString = format("%s  %d:%.2d", player, value / 60, math.mod(value, 60));
	else
		DMPrint("DamageMeters: Internal error.  Invalid quantity "..DamageMeters_quantity);
		DamageMeters_quantity = DamageMeters_Quantity_DAMAGE;
	end

	if (DamageMeters_colorScheme == 1 or DamageMeters_colorScheme == 2) then
		if (red == 1.00 and green == 1.00 and blue == 1.00) then
			red = 0.5 + 0.5 * (1.0 - pulseMag);
			green = red;
			blue = red;
		else
			red = pulseMag > red and pulseMag or red;
			green = pulseMag > green and pulseMag or green;
			blue = pulseMag > blue and pulseMag or blue;
		end
	else	
		blue = pulseMag;
	end

	-- This works: it is possible to color the text.
	--barString = "|cFF8F8FFF"..barString.."|r";

	DamageMeters_bars[barIndex]:Show();
	DamageMeters_bars[barIndex]:SetStatusBarColor(red, green, blue);
	DamageMeters_bars[barIndex]:SetValue(value);
	DamageMeters_text[barIndex]:SetText(barString);
	-- After 1300 patch text wouldn't appear without this.
	DamageMeters_text[barIndex]:Show();
end

-------------------------------------------------------------------------------

function DamageMeters_Clear(leave)
	local last = table.getn(DamageMeters_table);
	if (last == 0) then
		return;
	end

	local first = 1;
	if (leave ~= nil) then
		--DMPrint("leave = '"..leave.."'");
		local c = tonumber(leave);
		if (c) then
			first = leave + 1;
			if (first < 1) then
				first = 1;
			end
		end
	end

	DMPrint("DamageMeters: Removing "..first.." to "..last..".");
	local i;
	for i = last,first,-1 do
		if (DamageMeters_bars[i]) then
			DamageMeters_bars[i]:SetValue(0);
			DamageMeters_text[i]:SetText("");
		end

		table.remove(DamageMeters_table);
	end

	DMPrint("DamageMeters: "..table.getn(DamageMeters_table).." items remaining.");
end

function DamageMeters_Test(countArg)
	DamageMeters_Clear();

	local count = DamageMeters_BarCount;
	if (countArg) then
		count = tonumber(countArg);
		if (not count) then
			count = DamageMeters_BarCount
		end

		if (count > DamageMeters_BarCount) then
			count = DamageMeters_BarCount;
		end
	end

	local index;
	for index = 1,count do
		DamageMeters_AddDamage("Test"..index, 10*index, false, DamageMeters_Relation_FRIENDLY, true);
	end
end

function DamageMeters_GetPlayerIndex(player)
	local i;
	for i = 1, table.getn(DamageMeters_table) do
		if (DamageMeters_table[i].player == player) then
			return i;
		end
	end
	
	return nil;
end

function DamageMeters_Add(player)
	if (player) then
		DamageMeters_AddDamage(player, 1, false, DamageMeters_Relation_FRIENDLY);
	end
end

function DamageMeters_SetSort(sortArg)
	local usage = true;
	if (sortArg) then
		local sort = tonumber(sortArg);
		if (sort) then
			if (sort >= 1 and sort <= DamageMeters_Sort_MAX) then
				DamageMeters_sort = sort;
				DMPrint("DamageMeters: Setting sort to "..DamageMeters_sort);
				usage = false;
			else
				DMPrint("DamageMeters: Invalid sort "..sort..".");
			end
		end
	end

	if (usage) then
		DMPrint("DamageMeters: Current sort = "..getglobal("DamageMeters_Sort_STRING"..DamageMeters_sort));
		local i;
		for i=1,DamageMeters_Sort_MAX do
			DMPrint(" "..i..": "..getglobal("DamageMeters_Sort_STRING"..i));
		end
	end
end

function DamageMeters_SetQuantity(quantArg)
	local usage = true;
	if (quantArg) then
		local quant = tonumber(quantArg);
		if (quant) then
			if (quant >= 1 and quant <= DamageMeters_Quantity_MAX) then
				DamageMeters_quantity = quant;
				DMPrint("DamageMeters: Setting visible quantity to "..DamageMeters_quantity);
				usage = false;
			else
				DMPrint("DamageMeters: Invalid quantity "..quant..".");
			end
		end
	end

	if (usage) then
		DMPrint("DamageMeters: Current quantity = "..getglobal("DamageMeters_Quantity_STRING"..DamageMeters_quantity));
		local i;
		for i=1,DamageMeters_Quantity_MAX do
			DMPrint(" "..i..": "..getglobal("DamageMeters_Quantity_STRING"..i));
		end
	end

	DamageMeters_SetBackgroundColor();
end

function DamageMeters_SetBackgroundColor()
	local frame = getglobal("DamageMetersFrame");
	if (frame) then
		if (DamageMeters_quantity == DamageMeters_Quantity_DAMAGE) then
			frame:SetBackdropColor(0.3, 0.0, 0.0, 1);
		elseif (DamageMeters_quantity == DamageMeters_Quantity_HEALING) then
			frame:SetBackdropColor(0.0, 0.3, 0.0, 1);
		else
			frame:SetBackdropColor(0.0, 0.0, 0.5, 1);
		end
	end
end

function DamageMeters_AddDamage(player, amount, crit, relationship, silent, healing)
	local currentTime = GetTime();

	local index = DamageMeters_GetPlayerIndex(player);
	if (nil == index) then
		if (table.getn(DamageMeters_table) >= DamageMeters_BARCOUNT_MAX) then
			return;
		end

		index = table.getn(DamageMeters_table) + 1;
		table.setn(DamageMeters_table, index);
		DamageMeters_table[index] = {};
		DamageMeters_table[index].damage = 0;
		DamageMeters_table[index].healing = 0;
		DamageMeters_table[index].hitCount = 0;
		DamageMeters_table[index].critCount = 0;
		DamageMeters_table[index].player = player;
		DamageMeters_table[index].lastTime = currentTime;
		DamageMeters_table[index].relationship = relationship;
		--Print("relationship = "..relationship);

		if (relationship == DamageMeters_Relation_SELF) then
			DamageMeters_table[index].class = UnitClass("player");
			--Print("Adding self, class = "..DamageMeters_table[index].class);
		elseif (relationship == DamageMeters_Relation_PARTY) then
			local i;
			for i=1,5 do
				local partyUnitName = "party"..i;
				local partyName = UnitName(partyUnitName);
				if (partyName == player) then
					DamageMeters_table[index].class = UnitClass(partyUnitName);
					--Print("Party member found: index = "..i..", class = "..DamageMeters_table[index].class);
					i = 5
				end
			end
		else
			DamageMeters_UpdateRaidMemberClasses();
		end

		--if (not silent) then
		--	DMPrint("DamageMeters: Adding new player ["..index.."], "..player);
		--end

		if (index <= DamageMeters_BarCount) then
			DamageMeters_bars[index]:Show();
		end
	end

	if (not healing) then
		DamageMeters_table[index].damage = DamageMeters_table[index].damage + amount;
	else
		DamageMeters_table[index].healing = DamageMeters_table[index].healing + amount;
	end
	--! heal crits, etc
	DamageMeters_table[index].hitCount = DamageMeters_table[index].hitCount + 1;
	if (crit) then
		DamageMeters_table[index].critCount = DamageMeters_table[index].critCount + 1;
	end

	local idle = currentTime - DamageMeters_table[index].lastTime;
	DamageMeters_table[index].lastTime = currentTime;
end

function DamageMeters_Report(arg1)
	local destChar = "c";
	local count = table.getn(DamageMeters_table);
	local tellTarget = "";
	
	local argsParsed = false;

	if (arg1 == "") then
		argsParsed = true;
	end

	local a, b, c;
	if (not argsParsed) then
		for a, b, c in string.gfind(arg1, "(.)(%d) (.+)") do
			destChar = a;
			count = tonumber(b);
			tellTarget = c;
			--DMPrint(1);
			argsParsed = true;
		end
	end
	if (not argsParsed) then
		for a, c in string.gfind(arg1, "(.) (%d)") do
			destChar = a;
			--tellTarget = c;
			tellTarget = format("%d", c);
			--DMPrint(2);
			argsParsed = true;
		end
	end
	if (not argsParsed) then
		for a, b in string.gfind(arg1, "(.)(%d)") do
			destChar = a;
			count = tonumber(b);
			--DMPrint(3);
			argsParsed = true;
		end
	end
	if (not argsParsed) then
		for a, c in string.gfind(arg1, "(.) (.+)") do
			destChar = a;
			tellTarget = c;
			--DMPrint(4);
			argsParsed = true;
		end
	end
	if (not argsParsed) then
		for a in string.gfind(arg1, "(.)") do
			destChar = a;
			--DMPrint(5);
			argsParsed = true;
		end
	end
	--DMPrint("."..destChar.."."..count.."."..tellTarget..".");

	if (not argsParsed) then
		DMPrint("DamageMeters: Invalid arguments.");
		DMPrint(DamageMeters_Help_dmreport);
	end

	local destination;
	local invert = false;
	if (destChar) then
		--DMPrint("DamageMeters_Report("..arg1..")");

		local lowerDestChar = string.lower(destChar);
		if (lowerDestChar ~= destChar) then
			invert = true;
			destChar = lowerDestChar;
		end

		if (destChar == "c") then
			destination = "CONSOLE";
		elseif (destChar == "p") then
			destination = "PARTY";
		elseif (destChar == "s") then
			destination = "SAY";
		elseif (destChar == "r") then
			destination = "RAID";
		elseif (destChar == "w") then
			destination = "WHISPER";
		elseif (destChar == "h") then
			destination = "CHANNEL";
		elseif (destChar == "g") then
			destination = "GUILD";

		else
			DMPrint("DamageMeters: Invalid report target '"..destChar.."'.");
			return;
		end
	end

	if (destination == "WHISPER" and tellTarget == "") then
		DMPrint("DamageMeters: Whisper specified but no player given.");
		return;
	elseif (destination == "CHANNEL" and tellTarget == "") then
		DMPrint("DamageMeters: Channel specified but no number given.");
		return;
	end

	DamageMeters_DoReport(destination, invert, count, tellTarget);
end

function DamageMeters_DumpTable()
	DMPrint(table.getn(DamageMeters_table).." elements:");

	local index;
	local info;
	for index,info in DamageMeters_table do 
		DMPrint(index..": "..info.player);
	end
end

function DamageMeters_DoReport(destination, invert, count, tellTarget)
	local editBox = DEFAULT_CHAT_FRAME.editBox;
	local msg;

	msg = "Damage Meters: <"..getglobal("DamageMeters_Quantity_STRING"..DamageMeters_quantity).."> report on "..count.."/"..table.getn(DamageMeters_table).." sources:";
	DamageMeters_SendReportMsg(msg, destination, tellTarget);

	local start = 1;
	local finish = table.getn(DamageMeters_table);
	if (count < finish) then
		finish = count;
	end

	local step = 1;
	if (invert) then
		start = finish;
		finish = 1;
		step = -1;
	end

	local totalValue = 0;
	local index;
	local info;
	if (DamageMeters_Quantity_HEALING == DamageMeters_quantity) then
		for index,info in DamageMeters_table do 
			totalValue = totalValue + info.healing;
		end
	elseif (DamageMeters_Quantity_DAMAGE == DamageMeters_quantity) then
		for index,info in DamageMeters_table do 
			totalValue = totalValue + info.damage;
		end
	end

	local currentTime = GetTime();
	local i;
	for i = start, finish, step do
		local value;	
		if (DamageMeters_Quantity_HEALING == DamageMeters_quantity) then
			value = DamageMeters_table[i].healing;
			local percentage = (totalValue > 0) and (100 * value / totalValue) or 0;
			msg = format("#%.2d:  %.2f%%  %-16s  %d", i, percentage, DamageMeters_table[i].player, value);
		elseif (DamageMeters_Quantity_DAMAGE == DamageMeters_quantity) then
			value = DamageMeters_table[i].damage;
			local percentage = (totalValue > 0) and (100 * value / totalValue) or 0;
			msg = format("#%.2d:  %.2f%%  %-16s  %d", i, percentage, DamageMeters_table[i].player, value);
		else
			local idleTime = currentTime - DamageMeters_table[i].lastTime;
			msg = format("#%.2d: %-16s  %d:%.02d", i, DamageMeters_table[i].player, idleTime / 60, math.mod(idleTime, 60));
		end
		

		DamageMeters_SendReportMsg(msg, destination, tellTarget);
	end
end

function DamageMeters_SendReportMsg(msg, destination, tellTarget)
	local editBox = DEFAULT_CHAT_FRAME.editBox;
	if (destination == "CONSOLE") then
		DMPrint(msg);
	elseif (destination == "CHANNEL") then
		--DMPrint("Destination = "..destination..", tellTarget = "..tellTarget);
		SendChatMessage(msg, destination, editBox.language, GetChannelName(tellTarget));
	else
		SendChatMessage(msg, destination, editBox.language, tellTarget);
	end
end

function DamageMeters_SetCount(arg1)
	local count = 0;
	if (not arg1 or arg1 == "") then
		DMPrint("DamageMeters: No count argument specified, setting to max ("..DamageMeters_BARCOUNT_MAX..").");
		count = DamageMeters_BARCOUNT_MAX;
	else 
		count = tonumber(arg1);
		if (count > DamageMeters_BARCOUNT_MAX) then
			count = DamageMeters_BARCOUNT_MAX;
		end
	end

	if (not this) then
		DMPrint("DamageMeters_SetCount: this = nil");
	end

	DamageMeters_BarCount = count;
	DamageMetersFrame_GenerateFrame();
end

function DamageMeters_Reset()
	DamageMetersFrame_GenerateFrame();
end

function DamageMeters_ResetPos()
	local frame = getglobal("DamageMetersFrame");
	if (not frame) then
		DMPrint("DamageMetersFrame_Reset : Error getting frame.");
	end

	DMPrint("DamageMeters: Resetting frame position.");
	frame:ClearAllPoints();
	frame:SetPoint("RIGHT", "UIParent", "RIGHT", 0, 0);
end

function DamageMetersFrame_OnClick(arg1)
	DMPrint("Clear");
	--!
	DamageMeters_Clear();

	if ( arg1 == "RightButton" ) then
		DamageMeters_Clear();
	end
end

-- In the case of a tie, sort alphabetically.
function DamageMeters_DoSort()
	if (DamageMeters_quantity == DamageMeters_Quantity_DAMAGE) then
		table.sort(DamageMeters_table, 
				function(a,b) 
					local result;
					if (a.damage == b.damage) then
						result = a.player < b.player;
					else
						result = a.damage > b.damage;
					end

					if (DamageMeters_sort == DamageMeters_Sort_INCREASING) then
						result = not result;
					end
					return result;
				end
				);
	elseif (DamageMeters_quantity == DamageMeters_Quantity_HEALING) then
		table.sort(DamageMeters_table, 
				function(a,b) 
					local result;
					if (a.healing == b.healing) then
						result = a.player < b.player;
					else
						result = a.healing > b.healing; 
					end

					if (DamageMeters_sort == DamageMeters_Sort_INCREASING) then
						result = not result;
					end
					return result;
				end
				);
	elseif (DamageMeters_quantity == DamageMeters_Quantity_TIME) then
		table.sort(DamageMeters_table, 
				function(a,b) 
					local result;
					if (a.lastTime == b.lastTime) then
						result = a.player < b.player;
					else
						result = a.lastTime < b.lastTime;
					end

					if (DamageMeters_sort == DamageMeters_Sort_INCREASING) then
						result = not result;
					end
					return result;
				end
				);
	end
end

function DamageMeters_Show()
	local frame = getglobal("DamageMetersFrame");
	if (frame:IsVisible()) then
		frame:Hide();
	else
		frame:Show();
	end
end

function DamageMeters_Hide()
	getglobal("DamageMetersFrame"):Hide();
end

function DamageMetersFrame_OnEvent()
	local creatureName;
	local damage;
	local spell;
	local petName;
	local partialBlock;
	local damageType;

	if (event == "VARIABLES_LOADED") then
		DamageMeters_Reset();
		return;
	end

	if ( event == "RAID_ROSTER_UPDATE" ) then
		DamageMeters_UpdateRaidMemberClasses();
		return;
	end

	---------------------------

	if(	event == "CHAT_MSG_COMBAT_SELF_HITS" ) then
		--DMVerbose(event.." : "..arg1);
		for creatureName, damage in string.gfind(arg1, "You hit (.+) for (%d+)") do
			DamageMeters_AddDamage(UnitName("Player"), damage, false, DamageMeters_Relation_SELF);
			return;
		end
		for creatureName, damage in string.gfind(arg1, "You crit (.+) for (%d+)") do
			DamageMeters_AddDamage(UnitName("Player"), damage, true, DamageMeters_Relation_SELF);
			return;
		end
	end

	if ( event == "CHAT_MSG_SPELL_SELF_DAMAGE" )then
		--DMVerbose(event.." : "..arg1);
		for spell, creatureName, damage in string.gfind(arg1, "Your (.+) hits (.+) for (%d+).") do
			DamageMeters_AddDamage(UnitName("Player"), damage, false, DamageMeters_Relation_SELF);
			return;
		end
		for spell, creatureName, damage in string.gfind(arg1, "Your (.+) crits (.+) for (%d+).") do
			DamageMeters_AddDamage(UnitName("Player"), damage, true, DamageMeters_Relation_SELF);
			return;
		end
		for spell, creatureName, damage, partialBlock in string.gfind(arg1, "Your (.+) hits (.+) for (%d+). ((%d+) blocked)") do
			DamageMeters_AddDamage(UnitName("Player"), damage, false, DamageMeters_Relation_SELF);
			return;	
		end
	end

	if(	event == "CHAT_MSG_COMBAT_PET_HITS" ) then
		--DMVerbose(event.." : "..arg1);
		for petName, creatureName, damage in string.gfind(arg1, "(.+) hits (.+) for (%d+).") do
			DamageMeters_AddDamage(petName, damage, false, DamageMeters_Relation_PET);
			return;
		end
		
		for petName, creatureName, damage in string.gfind(arg1, "(.+) crits (.+) for (%d+).") do
			DamageMeters_AddDamage(petName, damage, true, DamageMeters_Relation_PET);
			return;
		end
	end

	if ( event == "CHAT_MSG_SPELL_PET_DAMAGE" ) then
		--DMVerbose(event.." : "..arg1);
		for petName, spell, creatureName, damage in string.gfind(arg1, "(.+)'s (.+) hits (.+) for (%d+).") do
			DamageMeters_AddDamage(petName, damage, false, DamageMeters_Relation_PET);
			return;
		end
		for petName, spell, creatureName, damage in string.gfind(arg1, "(.+)'s (.+) crits (.+) for (%d+).") do
			DamageMeters_AddDamage(petName, damage, true, DamageMeters_Relation_PET);
			return;
		end	
		for petName, spell, creatureName, damage in string.gfind(arg1, "(.+) casts (.+) on (.+) for (%d+).") do
			DamageMeters_AddDamage(petName, damage, false, DamageMeters_Relation_PET);
			return;
		end
		for petName, spell, creatureName, damage, partialBlock in string.gfind(arg1, "(.+)'s (.+) hits (.+) for (%d+). ((%d+) blocked)") do
			DamageMeters_AddDamage(petName, damage, false, DamageMeters_Relation_PET);
			return;	
		end		
	end

	-- TODO: Figure out this mess.
	if ( event == "CHAT_MSG_SPELL_PERIODIC_CREATURE_DAMAGE" or event == "CHAT_MSG_SPELL_PERIODIC_HOSTILEPLAYER_DAMAGE" ) then
		--DMVerbose(event.." : "..arg1);
		
		-- Below code isnt finished because you cant tell who cast the spell
		
		for creatureName, spell in string.gfind(arg1, "(.+) is afflicted by (.+).") do	
			return;
		end
		
		for spell in string.gfind(arg1, "You are afflicted by (.+).") do	
			return;
		end
	
		-- todo: this isn't checking for the filter, thus friendly other faction's dots can 
		-- slip through
		
		for creatureName, damage, damageType, spell in string.gfind(arg1, "(.+) suffers (%d+) (.+) damage from your (.+).") do
			DamageMeters_AddDamage(UnitName("Player"), damage, false, DamageMeters_Relation_SELF);
			return;
		end		
	
		for creatureName, damage, damageType, playerName, spell in string.gfind(arg1, "(.+) suffers (%d+) (.+) damage from (.+)'s (.+).") do
			if (creatureName == playerName) then 
				return;	
			end
			
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_FRIENDLY);
			return;
		end	
	end

	if (event == "CHAT_MSG_COMBAT_PARTY_HITS") then
		--DMVerbose(event.." : "..arg1);
		for playerName, creatureName, damage in string.gfind(arg1, "(.+) hits (.+) for (%d+).") do
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_PARTY);
			return;
		end
		for playerName, creatureName, damage in string.gfind(arg1, "(.+) crits (.+) for (%d+).") do
			DamageMeters_AddDamage(playerName, damage, true, DamageMeters_Relation_PARTY);
			return;
		end
	end

	if (event == "CHAT_MSG_COMBAT_FRIENDLYPLAYER_HITS") then
		--DMVerbose(event.." : "..arg1);
		for playerName, creatureName, damage in string.gfind(arg1, "(.+) hits (.+) for (%d+).") do
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_PARTY);
			return;
		end
		for playerName, creatureName, damage in string.gfind(arg1, "(.+) crits (.+) for (%d+).") do
			DamageMeters_AddDamage(playerName, damage, true, DamageMeters_Relation_PARTY);
			return;
		end
	end

	if (event == "CHAT_MSG_SPELL_PARTY_DAMAGE") then
		--DMVerbose(event.." : "..arg1);
		for playerName, spell, creatureName, damage in string.gfind(arg1, "(.+)'s (.+) hits (.+) for (%d+).") do		
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_PARTY);
			return;
		end
		for playerName, spell, creatureName, damage in string.gfind(arg1, "(.+)'s (.+) crits (.+) for (%d+).") do
			DamageMeters_AddDamage(playerName, damage, true, DamageMeters_Relation_PARTY);
			return;
		end
		for playerName, spell, creatureName, damage, partialBlock in string.gfind(arg1, "(.+)'s (.+) hits (.+) for (%d+). ((%d+) blocked)") do
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_PARTY);
			return;
		end
	end

	if ((event == "CHAT_MSG_SPELL_FRIENDLYPLAYER_DAMAGE")) then
		--DMVerbose(event.." : "..arg1);
		for playerName, spell, creatureName, damage in string.gfind(arg1, "(.+)'s (.+) hits (.+) for (%d+).") do		
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_FRIENDLY);
			return;
		end
		for playerName, spell, creatureName, damage in string.gfind(arg1, "(.+)'s (.+) crits (.+) for (%d+).") do
			DamageMeters_AddDamage(playerName, damage, true, DamageMeters_Relation_FRIENDLY);
			return;
		end
		for playerName, spell, creatureName, damage, partialBlock in string.gfind(arg1, "(.+)'s (.+) hits (.+) for (%d+). ((%d+) blocked)") do
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_FRIENDLY);
			return;
		end
	end

	-- Damage Shields --

	if (event == "CHAT_MSG_SPELL_DAMAGESHIELDS_ON_SELF") then
		--DMVerbose(event.." : "..arg1);
		for damage, damageType, creatureName in string.gfind(arg1, "You reflect (%d+) (.+) to (.+).") do
			DamageMeters_AddDamage(UnitName("Player"), damage, false, DamageMeters_Relation_SELF);
			return;
		end
	end

	if (event == "CHAT_MSG_SPELL_DAMAGESHIELDS_ON_OTHERS") then
		--DMVerbose(event.." : "..arg1);
		for playerName, damage, damageType, creatureName in string.gfind(arg1, "(.+) reflects (%d+) (.+) to (.+).") do
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_FRIENDLY);
			return;
		end
	end

	--------------------------------

	if (event == "CHAT_MSG_SPELL_SELF_BUFF" or
		event == "CHAT_MSG_SPELL_PARTY_BUFF" or
		event == "CHAT_MSG_SPELL_FRIENDLYPLAYER_BUFF" or
		event == "CHAT_MSG_SPELL_HOSTILEPLAYER_BUFF") then
		for spell, creatureName, damage in string.gfind(arg1, "Your (.+) critically heals (.+) for (%d+).") do
			DM_CountMsg("Self Crit Heal");
			DamageMeters_AddDamage(UnitName("Player"), damage, false, DamageMeters_Relation_SELF, nil, true);
			return;
		end
		for spell, creatureName, damage in string.gfind(arg1, "Your (.+) heals (.+) for (%d+).") do
			DM_CountMsg("Self Heal");
			DamageMeters_AddDamage(UnitName("Player"), damage, false, DamageMeters_Relation_SELF, nil, true);
			return;
		end
		for playerName, spell, creatureName, damage in string.gfind(arg1, "(.+)'s (.+) critically heals (.+) for (%d+).") do
			DM_CountMsg("Other's Critical Heal");
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_FRIENDLY, nil, true);
			return;
		end
		
		for playerName, spell, creatureName, damage in string.gfind(arg1, "(.+)'s (.+) heals (.+) for (%d+).") do
			DM_CountMsg("Other's Heal");
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_FRIENDLY, nil, true);
			return;
		end
		local spell2;
		for playerName, spell, spell2, creatureName, damage in string.gfind(arg1, "(.+)'s (.+) (.+) heals (.+) for (%d+).") do
			DM_CountMsg("Other's Two-word Heal");
				DMPrintD(arg1.." 1='"..spell.."', 2= '"..spell2.."'");
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_FRIENDLY, nil, true);
			return;
		end
	end

	if (event == "CHAT_MSG_SPELL_PERIODIC_SELF_BUFFS" or
		event == "CHAT_MSG_SPELL_PERIODIC_PARTY_BUFFS" or
		event == "CHAT_MSG_SPELL_PERIODIC_FRIENDLYPLAYER_BUFFS" or
		event == "CHAT_MSG_SPELL_PERIODIC_HOSTILEPLAYER_BUFFS") then
		for damage, playerName, spell in string.gfind(arg1, "You gain (%d+) health from (.+)'s (.+).") do
			DM_CountMsg("Other's Hot on Self");
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_FRIENDLY, nil, true);
			return;
		end
		for damage, spell in string.gfind(arg1, "You gain (%d+) health from (.+).") do
			DM_CountMsg("Self Hot");
			DamageMeters_AddDamage(UnitName("Player"), damage, false, DamageMeters_Relation_SELF, nil, true);
			return;
		end
		for creatureName, damage, spell in string.gfind(arg1, "(.+) gains (%d+) health from your (.+).") do		
			DM_CountMsg("Self Hot on Others");
			DamageMeters_AddDamage(UnitName("Player"), damage, false, DamageMeters_Relation_SELF, nil, true);
			return;
		end
		for creatureName, damage, playerName, spell in string.gfind(arg1, "(.+) gains (%d+) health from (.+)'s (.+).") do
			DM_CountMsg("Other's Hot on Others");
			DamageMeters_AddDamage(playerName, damage, false, DamageMeters_Relation_FRIENDLY, nil, true);
			return;
		end
	end

	-- cyn's heals are "hostile players buff spells"

	local sub = string.sub(event, 1, 8);
	--DMPrintD(event.." fell through, sub = "..sub);
	if (sub == "CHAT_MSG") then
		--DMPrintD(event.." fell through");
		if (string.find(arg1, "heals") or
			string.find(arg1, "damage") or
			--string.find(arg1, "gain") or
			string.find(arg1, "health") or
			string.find(arg1, "reflect") or
			string.find(arg1, "hits")) then
			if ((not string.find(arg1, "falls")) and
				(not string.find(arg1, "fall")) and
				(not string.find(arg1, "fire damage"))) then
				DMPrintD("Unparsed: "..event..":");
				DMPrintD(arg1);
			end
		end
	end
	--! X suffers # points of fire damage.
end

function DM_CountMsg(desc)
	if (DamageMeters_msgCounts[desc]) then
		DamageMeters_msgCounts[desc] = DamageMeters_msgCounts[desc] + 1;
	else
		DamageMeters_msgCounts[desc] = 1;
	end
	--DMPrintD("Caught '"..desc.."', count = "..DamageMeters_msgCounts[desc]);
end

function DM_DumpMsg()
	local i;
	local desc;
	for key,count in DamageMeters_msgCounts do
		DMPrintD(key.." : "..count)
	end;
end

-------------------------------------------------------------------------------

function DamageMetersBarTemplate_OnEnter()
	if (not this:IsVisible()) then
		return;
	end

	local index;
	local info;
	local totalDamage = 0;
	local totalHealing = 0;
	for index,info in DamageMeters_table do 
		totalDamage = totalDamage + info.damage; 
		totalHealing = totalHealing + info.healing; 
	end

	DamageMeters_tooltipBarIndex = this:GetID();

	DamageMetersTooltip:SetOwner(DamageMetersFrame, "ANCHOR_LEFT");
	DamageMeters_SetTooltipText(totalDamage, totalHealing);

	DamageMetersTooltip:ClearAllPoints();
	local anchorPoint = "RIGHT";
	local relativePoint = "LEFT";
	local xOffset = -4;

	local x,y = DamageMetersFrame:GetCenter();
	local screenWidth = UIParent:GetWidth();
	if (x~=nil and screenWidth~=nil) then
		if (x < (screenWidth / 2)) then
			anchorPoint = "LEFT";
			relativePoint = "RIGHT";
			xOffset = 4;
		end
	end

	DamageMetersTooltip:SetPoint(anchorPoint, this:GetName(), relativePoint, xOffset, 0);
	
	-- added 2005-01-13 by arys
	-- makes clock tooltip scale correctly
	DamageMetersTooltip:SetScale(this:GetScale());
end

function DamageMetersBarTemplate_OnLeave()
	DamageMeters_tooltipBarIndex = nil;
	DamageMetersTooltip:Hide();
end

function DamageMetersBarTemplate_OnClick()
	local index = this:GetID();
	
	--Print("Click '"..arg1.."'");
	if ( button == "LeftButton" ) then
		if (index <= table.getn(DamageMeters_table)) then
			--Print("Targetting "..DamageMeters_table[index].player);
			TargetByName(DamageMeters_table[index].player);
		end
	end
end

function DamageMeters_SetTooltipText(totalDamage, totalHealing)
	local struct = DamageMeters_table[DamageMeters_tooltipBarIndex];

	if ((nil == struct) or (not DamageMeters_tooltipBarIndex) or (DamageMeters_tooltipBarIndex > table.getn(DamageMeters_table))) then
		-- (nil == struct.player) or 
		DamageMeters_tooltipBarIndex = nil;
		DamageMetersTooltip:Hide();
		return;
	end

	DamageMetersTooltip:SetText(struct.player, 1.00, 1.00, 1.00);
	
	local damagePercent = (totalDamage > 0) and (100 * struct.damage / totalDamage) or 0;
	local healingPercent = (totalHealing > 0) and (100 * struct.healing / totalHealing) or 0;

	text = format("Damage = %d (%.2f%% of total)\nDamage Count = %d\nCrits = %d (%.2f%%*)\nHealing = %d (%.2f%% of total)\nTime since last action = %.1fs\nRelationship = %s", 
			struct.damage, 
			damagePercent,
			struct.hitCount,
			struct.critCount,
			struct.hitCount and (struct.critCount / struct.hitCount) * 100 or 0,
			struct.healing,
			healingPercent,
			GetTime() - struct.lastTime,
			getglobal("DamageMeters_Relation_STRING"..struct.relationship));
	if (struct.class) then
		text = text.."\nClass = "..struct.class;
	end

	DamageMetersTooltip:AddLine(text, NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b, 1);
	DamageMetersTooltip:Show();
end

function DamageMeters_Help()
	local i;
	for i = 1, DamageMeters_COMMANDCOUNT do
		DMPrint(getglobal("DamageMeters_Help_FULLDESC"..i));
	end
end

function DamageMeters_Save()
	DamageMeters_savedTable = {};

	DMPrint("DamageMeters: Saving table.");
	local index,struct;
	for index,struct in DamageMeters_table do
		DamageMeters_savedTable[index] = DamageMeters_table[index];
	end
end

function DamageMeters_Restore()
	if (DamageMeters_savedTable and (table.getn(DamageMeters_savedTable) > 0)) then
		DMPrint("DamageMeters: Restoring saved table.");
		DamageMeters_table = DamageMeters_savedTable;
		DamageMeters_savedTable = {};
	else
		DMPrint("DamageMeters: No saved table.");
	end
end

function DamageMeters_Merge()
	DMPrint("DamageMeters: Merging saved table with current.");

	local index,struct;
	for index,struct in DamageMeters_savedTable do
		local player = struct.player;

		local existingIndex = DamageMeters_GetPlayerIndex(player);
		if (nil == existingIndex) then
			local targetLength = table.getn(DamageMeters_table);
			if (targetLength < DamageMeters_BARCOUNT_MAX) then
				-- Copy entire entry over.
				--DMPrint("  "..player.." not found, creating new entry.");
				local newIndex = targetLength + 1;
				DamageMeters_table[newIndex] = struct;
			end
		else
			-- Merge with existing data.
			--DMPrint("  "..player.." found, merging data.");
			local target = DamageMeters_table[existingIndex];

			target.damage = target.damage + struct.damage;
			target.healing = target.healing + struct.healing;
			target.critCount = target.critCount + struct.critCount;
			target.hitCount = target.hitCount + struct.hitCount;
			target.lastTime = (target.lastTime > struct.lastTime) and target.lastTime or struct.lastTime;
			-- Leave relationship.
			if (not target.class) then
				target.class = struct.class;
			end
		end
	end
end

function DamageMeters_Swap()
	local tempTable = {};

	DMPrint("DamageMeters: Swapping normal ("..table.getn(DamageMeters_table)..") and saved ("..table.getn(DamageMeters_savedTable)..") table.");

	-- Save table to temp.
	local index,struct;
	for index,struct in DamageMeters_table do
		tempTable[index] = struct;
	end

	-- Copy saved to table.
	DamageMeters_table = {};
	for index,struct in DamageMeters_savedTable do
		DamageMeters_table[index] = struct;
	end

	-- Copy temp to saved.
	DamageMeters_savedTable = {};
	for index,struct in tempTable do
		DamageMeters_savedTable[index] = struct;
	end

	-- Regen frame to hide any buttons now not needed.
	DamageMetersFrame_GenerateFrame();
end

function DamageMetersFrame_OnClick(button)
	if ( button == "RightButton" ) then
		ToggleDropDownMenu(1, nil, getglobal(this:GetName().."DropDown"), this:GetName(), 0, 0);
		return;
	end
end

function DamageMeters_SetTextOptions(arg1)
	if (arg1 == "") then
		DMPrint("DamageMeters: Missing argument.");
		DMPrint(DamageMeters_Help_dmtext);
		return;
	end

	DamageMeters_showName = false;
	DamageMeters_showPercentage = false;
	DamageMeters_showValue = false;		

	if (arg1 == "0") then
		return;	
	end

	local i;
	for i = 1, string.len(arg1) do
		local char = string.lower(string.sub(arg1, i, i));

		if (char == "n") then
			DamageMeters_showName = true;
		elseif (char == "p") then
			DamageMeters_showPercentage = true;
		elseif (char == "v") then
			DamageMeters_showValue = true;
		end
	end
end

function DamageMeters_SetColorScheme(arg1)
	local showUsage = false;
	local colorScheme;
	if (not arg1 or arg1 == "") then
		showUsage = true;
	else
		colorScheme = tonumber(arg1);
		if (colorScheme < 1 or colorScheme > DamageMeters_colorScheme_MAX) then
			showUsage = true;
		end	
	end
	
	if (showUsage) then
		DMPrint(DamageMeters_Help_dmcolor);

		local i;
		for i=1, DamageMeters_colorScheme_MAX do
			DMPrint(i..": "..getglobal("DamageMeters_colorScheme_STRING"..i));
		end

		return;
	end

	DMPrint("DamageMeters: Setting color scheme to "..getglobal("DamageMeters_colorScheme_STRING"..colorScheme));
	DamageMeters_colorScheme = colorScheme;
end

function DamageMeters_UpdateRaidMemberClasses()
	local numRaidMembers = GetNumRaidMembers();
	local name, rank, subgroup, level, class, fileName, zone, online, isDead;
	local i;
	for i = 1,numRaidMembers do
		name, rank, subgroup, level, class, fileName, zone, online, isDead = GetRaidRosterInfo(i);

		local tableIndex = DamageMeters_GetPlayerIndex(name);
		if (tableIndex) then
			DamageMeters_table[tableIndex].class = class;
		end
	end
end
